#include "fsck.h"

static void reclaim_directory_page(struct nvfs_dir_head *ndh)
{
	struct nvfs_dir_entry *nde, *end;
	uint64_t dir_ino;
	struct nvfs_inode *dir_inode;

	// TODO: add these entries to lost+found if the inode is invalid

	dir_ino = le64_to_cpu(ndh->inode);
	if (dir_ino >= n_groups << log2_inodes_per_group)
		return;
	if (is_inode_reserved(dir_ino))
		return;
	if (!bmp_op(dir_ino, B_READ | B_INODE | B_DEV))
		return;
	dir_inode = p_inode(dir_ino);
	if (!S_ISDIR(le32_to_cpu(dir_inode->i_mode)))
		return;

	nde = ndh->nde;
	end = (void *)ndh + (1U << sb->log2_page_size);
	do {
		uint32_t namelen = nvfs_dir_entry_namelen(nde);
		size_t size = nvfs_dir_entry_needed_size(namelen);
		unsigned char type = nvfs_dir_entry_type(nde);
		if ((size_t)((void *)end - (void *)nde) < size)
			break;
		if (type) {
			if (memchr(nde->name, 0, namelen) ||
			    memchr(nde->name, '/', namelen))
				goto next;
			if (nvfs_find_de(dir_ino, (const char *)nde->name, namelen)) {
				static struct qkey q;
				fsck_query(&q, Q_JUST_PRINT, "the file '%.*s' already exists in target directory %"PRIu64"", NULL, (int)namelen, nde->name, dir_ino);
				goto next;
			}
			{
				static struct qkey q;
				fsck_query(&q, Q_JUST_PRINT, "reclaiming the file '%.*s' for inode %"PRIu64"", NULL, (int)namelen, nde->name, dir_ino);
			}
			add_directory_entry(dir_ino, (const char *)nde->name, namelen, le64_to_cpu(nde->inode), type);
		}
next:
		nde = (void *)nde + size;
	} while (nde < end);
}

int check_lost_blocks(void)
{
	uint64_t b;
	uint64_t global_count = 0;

	fsck_log("checking for lost blocks...");

#pragma omp parallel for
	for (b = 0; b < n_groups << log2_pages_per_group; b++) {
		bool bm, bd;

		if (!(b % BLOCK_PROGRESS_STEP)) {
#pragma omp atomic
			global_count += BLOCK_PROGRESS_STEP;
			fsck_progress(2, n_groups << log2_pages_per_group, global_count);
		}

		bm = bmp_op(b, B_READ | B_BLOCK | B_MEM);
		bd = bmp_op(b, B_READ | B_BLOCK | B_DEV);
		if (!bm && bd) {
			struct nvfs_dir_head *ndh = ptr + (b << sb->log2_page_size);
			if (ndh->magic == NVFS_DIR_HEAD_MAGIC &&
			    le64_to_cpu(ndh->self) == b << sb->log2_page_size) {
				static struct qkey q;
				if (fsck_query(&q, Q_PREEN, "lost directory page %"PRIu64"", "reclaim?", b)) {
					reclaim_directory_page(ndh);
					bmp_op(b, B_CLEAR | B_BLOCK | B_DEV);
					bd = false;
				}
			} else {
				static struct qkey q;
				if (fsck_query(&q, Q_PREEN, "lost block %"PRIu64"", "clear?", b)) {
					bmp_op(b, B_CLEAR | B_BLOCK | B_DEV);
					bd = false;
				}
			}
		}
		/*
			this check is already done in verify_block
		if (bm && !bd) {
			static struct qkey q;
			if (fsck_query(&q, Q_NORMAL, "the block %"PRIu64" is marked as clean on the device", "set?", b)) {
				bmp_op(b, B_CLEAR | B_BLOCK | B_SET);
				bd = true;
			}
		}
		*/
	}

	return 0;
}
